from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, Callable

import torch

from nebullvm.base import ModelParams, DeepLearningFramework, QuantizationType
from nebullvm.converters import ONNXConverter
from nebullvm.inference_learners.deepsparse import (
    DEEPSPARSE_INFERENCE_LEARNERS,
    DeepSparseInferenceLearner,
)
from nebullvm.optimizers import BaseOptimizer
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager
from nebullvm.utils.onnx import (
    get_input_names,
    get_output_names,
)


class DeepSparseOptimizer(BaseOptimizer):
    def optimize(
        self,
        model: torch.nn.Module,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        metric: Callable = None,
        input_data: DataManager = None,
    ) -> Optional[DeepSparseInferenceLearner]:
        if quantization_type is not None:
            return

        with TemporaryDirectory() as tmp_dir:
            converter = ONNXConverter()
            onnx_pruned_path = Path(tmp_dir) / "model_pruned.onnx"
            converter.convert(
                model, model_params, onnx_pruned_path, input_data
            )

            learner = DEEPSPARSE_INFERENCE_LEARNERS[output_library](
                input_tfms=input_tfms,
                network_parameters=model_params,
                onnx_path=onnx_pruned_path,
                input_names=get_input_names(str(onnx_pruned_path)),
                output_names=get_output_names(str(onnx_pruned_path)),
            )
        return learner



from tempfile import TemporaryDirectory

import pytest
import torch

from nebullvm.base import DeepLearningFramework
from nebullvm.inference_learners.tensor_rt import (
    PytorchTensorRTInferenceLearner,
)
from nebullvm.inference_learners.tensor_rt import NVIDIA_INFERENCE_LEARNERS
from nebullvm.optimizers import TensorRTOptimizer
from nebullvm.optimizers.tests.utils import get_onnx_model, get_torch_model


@pytest.mark.parametrize(
    ("output_library", "dynamic"),
    [
        (DeepLearningFramework.PYTORCH, True),
        (DeepLearningFramework.PYTORCH, False),
    ],
)
def test_tensor_rt(output_library: DeepLearningFramework, dynamic: bool):
    if not torch.cuda.is_available():
        # no need of testing the tensor rt optimizer on devices not
        # supporting CUDA.
        return
    with TemporaryDirectory() as tmp_dir:
        model_path, model_params = get_onnx_model(tmp_dir, dynamic)
        optimizer = TensorRTOptimizer()
        model = optimizer.optimize(model_path, output_library, model_params)
        assert isinstance(model, NVIDIA_INFERENCE_LEARNERS[output_library])

        # Test save and load functions
        model.save(tmp_dir)
        loaded_model = NVIDIA_INFERENCE_LEARNERS[output_library].load(tmp_dir)
        assert isinstance(
            loaded_model, NVIDIA_INFERENCE_LEARNERS[output_library]
        )

        inputs_example = list(model.get_inputs_example())
        res = model.predict(*inputs_example)
        assert res is not None

        if dynamic:
            inputs_example = [
                input_[: len(input_) // 2] for input_ in inputs_example
            ]
            res = model.predict(*inputs_example)
            assert res is not None


@pytest.mark.parametrize(
    ("dynamic",),
    [
        (True,),
        (False,),
    ],
)
def test_torch_tensor_rt(dynamic: bool):
    if not torch.cuda.is_available():
        # no need of testing the tensor rt optimizer on devices not
        # supporting CUDA.
        return
    with TemporaryDirectory() as tmp_dir:
        model_path, model_params = get_torch_model(dynamic)
        optimizer = TensorRTOptimizer()
        model = optimizer.optimize_from_torch(model_path, model_params)
        assert isinstance(model, PytorchTensorRTInferenceLearner)

        # Test save and load functions
        model.save(tmp_dir)
        loaded_model = PytorchTensorRTInferenceLearner.load(tmp_dir)
        assert isinstance(loaded_model, PytorchTensorRTInferenceLearner)

        inputs_example = list(model.get_inputs_example())
        res = model.predict(*inputs_example)
        assert res is not None

        if dynamic:  # Check also with a smaller bath_size
            inputs_example = [
                input_[: len(input_) // 2] for input_ in inputs_example
            ]
            res = model.predict(*inputs_example)
            assert res is not None



import cpuinfo
import torch

from nebullvm.base import ModelCompiler


def tvm_is_available() -> bool:
    try:
        import tvm  # noqa F401

        return True
    except ImportError:
        return False


def bladedisc_is_available() -> bool:
    try:
        import torch_blade  # noqa F401

        return True
    except ImportError:
        return False


def torch_tensorrt_is_available() -> bool:
    try:
        import torch_tensorrt  # noqa F401

        return True
    except ImportError:
        return False


def deepsparse_is_available() -> bool:
    try:
        import deepsarse  # noqa F401
    except ImportError:
        return False
    else:
        return True


def select_compilers_from_hardware_onnx():
    compilers = [ModelCompiler.ONNX_RUNTIME]
    if tvm_is_available():
        compilers.append(ModelCompiler.APACHE_TVM)
    if torch.cuda.is_available():
        compilers.append(ModelCompiler.TENSOR_RT)
    cpu_raw_info = cpuinfo.get_cpu_info()["brand_raw"].lower()
    if "intel" in cpu_raw_info:
        compilers.append(ModelCompiler.OPENVINO)
    return compilers



import json
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Callable, Tuple, Optional, Dict

import numpy as np
import torch
import torch.fx

from nebullvm.compressors.base import BaseCompressor
from nebullvm.utils.data import DataManager
from nebullvm.utils.venv import run_in_different_venv

FX_MODULE_NAME = "NebullvmFxModule"


def _save_with_torch_fx(model: torch.nn.Module, path: Path):
    traced_model = torch.fx.symbolic_trace(model)
    traced_model.to_folder(path, FX_MODULE_NAME)


def _load_with_torch_fx(path: Path):
    module_file = path / "module.py"
    with open(module_file, "r") as f:
        module_str = f.read()
    exec(module_str)
    model = eval(FX_MODULE_NAME)()
    model.load_state_dict(torch.load(path / "pruned_state_dict.pt"))
    return model


def _save_model(model: torch.nn.Module, path: Path):
    try:
        _save_with_torch_fx(model, path)
    except Exception:
        torch.save(model, path / "model.pt")
        return path / "model.pt"
    else:
        return path


def _load_model(path: Path):
    if path.is_file():
        return torch.load(path)
    else:
        return _load_with_torch_fx(path)


def _save_dataset(input_data: DataManager, path: Path):
    path.mkdir(exist_ok=True)
    for i, x in enumerate(input_data):
        torch.save(x, path / f"input_{i}.pt")


def _save_json(dictionary: Dict, path: Path):
    with open(path, "w") as f:
        json.dump(dictionary, f)


def _write_requirements_file(path: Path):
    requirements = "torch<=1.9\ntorchvision<=0.10\nsparseml\nsparsify\ntqdm"
    with open(path, "w") as f:
        f.write(requirements)


class SparseMLCompressor(BaseCompressor):
    def compress(
        self,
        model: torch.nn.Module,
        train_input_data: DataManager,
        eval_input_data: DataManager,
        metric_drop_ths: float,
        metric: Callable,
    ) -> Tuple[Any, Optional[float]]:
        script_path = (
            Path(__file__).parent / "scripts/neural_magic_training.py"
        )
        with TemporaryDirectory(dir=".") as tmp_dir:
            tmp_dir = Path(tmp_dir)
            requirements_file = tmp_dir / "requirements.txt"
            model_path = _save_model(model, tmp_dir)
            training_data_dir = tmp_dir / "train"
            eval_data_dir = tmp_dir / "eval"
            config_file = tmp_dir / "config.json"
            pruned_model_path = (
                tmp_dir / "pruned_model.pt"
                if model_path.is_file()
                else tmp_dir
            )

            _write_requirements_file(requirements_file)
            _save_dataset(train_input_data, training_data_dir)
            _save_dataset(eval_input_data, eval_data_dir)
            _save_json(self._config, config_file)

            run_in_different_venv(
                str(requirements_file),
                str(script_path),
                "--model",
                f"{model_path}",
                "--train_dir",
                f"{training_data_dir}",
                "--eval_dir",
                f"{eval_data_dir}",
                "--config",
                f"{config_file}",
                "--pruned_model",
                f"{pruned_model_path}",
            )

            pruned_model = _load_model(pruned_model_path)

            error = self._compute_error(
                model, pruned_model, eval_input_data, metric
            )
            if error > metric_drop_ths:
                return None, None
            new_metric_ths = metric_drop_ths - error

        return pruned_model, new_metric_ths

    @staticmethod
    @torch.no_grad()
    def _compute_error(
        model: torch.nn.Module,
        pruned_model: torch.nn.Module,
        eval_input_data: DataManager,
        metric: Callable,
    ) -> float:
        if len(eval_input_data) == 0:
            return np.inf
        metric_val = 0.0
        model.eval()
        pruned_model.eval()
        for inputs, y in eval_input_data:
            model_pred = model(*inputs)
            pruned_pred = pruned_model(*inputs)
            metric_val += metric(model_pred, pruned_pred, y)
        return metric_val / len(eval_input_data)

    @staticmethod
    def _get_default_config() -> Dict:
        return {
            "training_epochs": 10,
            "epochs_pruning_window": {"start_epoch": 0, "end_epoch": 10},
            "loss_fn": "CrossEntropy",
            "lr": 1e-3,
            "momentum": 0.9,
        }

    @property
    def config_key(self) -> str:
        return "sparseml"



from packaging import version
from types import ModuleType


def check_module_version(
    module: ModuleType, min_version: str = None, max_version: str = None
) -> bool:
    installed_version = module.__version__

    if min_version is not None:
        if version.parse(installed_version) < version.parse(min_version):
            return False

    if max_version is not None:
        if version.parse(installed_version) > version.parse(max_version):
            return False

    return True



from tempfile import TemporaryDirectory

import cpuinfo
import pytest

from nebullvm.base import DeepLearningFramework
from nebullvm.inference_learners.openvino import (
    OPENVINO_INFERENCE_LEARNERS,
)
from nebullvm.optimizers.openvino import OpenVinoOptimizer
from nebullvm.optimizers.tests.utils import get_onnx_model


@pytest.mark.parametrize(
    ("output_library", "dynamic"),
    [
        (DeepLearningFramework.PYTORCH, True),
        (DeepLearningFramework.PYTORCH, False),
    ],
)
def test_openvino(output_library: DeepLearningFramework, dynamic: bool):
    if "intel" not in cpuinfo.get_cpu_info()["brand_raw"].lower():
        # No intel cpu detected
        return
    with TemporaryDirectory() as tmp_dir:
        model_path, model_params = get_onnx_model(tmp_dir, dynamic)
        optimizer = OpenVinoOptimizer()
        model = optimizer.optimize(model_path, output_library, model_params)
        assert isinstance(model, OPENVINO_INFERENCE_LEARNERS[output_library])

        # Test save and load functions
        model.save(tmp_dir)
        loaded_model = OPENVINO_INFERENCE_LEARNERS[output_library].load(
            tmp_dir
        )
        assert isinstance(
            loaded_model, OPENVINO_INFERENCE_LEARNERS[output_library]
        )

        inputs_example = list(model.get_inputs_example())
        res = model.predict(*inputs_example)
        assert res is not None

        if dynamic:  # Check also with a smaller bath_size
            inputs_example = [
                input_[: len(input_) // 2] for input_ in inputs_example
            ]
            res = model.predict(*inputs_example)
            assert res is not None



import json
import os
import platform
import select
import sys
import uuid
from pathlib import Path
from typing import Any, Optional

import psutil
import requests
import torch.cuda
from cpuinfo import cpuinfo

from nebullvm.base import (
    DeepLearningFramework,
    ModelCompiler,
    QuantizationType,
)
from nebullvm.config import VERSION

NEBULLVM_METADATA_PATH = Path.home() / ".nebullvm/collect.json"


def _input_with_timeout(message: str, timeout: int):
    print(message)
    i, o, e = select.select([sys.stdin], [], [], timeout)
    if i:
        input_str = sys.stdin.readline().strip()
    else:
        raise TimeoutError()
    return input_str


def _read_model_size(model: Any):
    if isinstance(model, str) or isinstance(model, Path):
        size = os.stat(str(model)).st_size
    elif isinstance(model, torch.nn.Module):
        size = sum(
            param.nelement() * param.element_size()
            for param in model.parameters()
        )
    else:
        # we assume it is a tf_model
        size = model.count_params() * 4  # assuming full precision 32 bit
    return f"{round(size * 1e-6, 2)} MB"


class FeedbackCollector:
    def __init__(self):
        if NEBULLVM_METADATA_PATH.exists():
            with open(NEBULLVM_METADATA_PATH, "r") as f:
                nebullvm_metadata = json.load(f)
        else:
            nebullvm_metadata = self._ask_permissions()
            NEBULLVM_METADATA_PATH.parent.mkdir(exist_ok=True)
            with open(NEBULLVM_METADATA_PATH, "w") as f:
                json.dump(nebullvm_metadata, f)
        self._is_active = nebullvm_metadata["allow_feedback_collection"]
        self._model_info = None
        self._model_id = None
        self._latency_dict = None
        self._hw_info = {
            "cpu": cpuinfo.get_cpu_info()["brand_raw"],
            "operative_system": platform.system(),
            "ram": f"{round(psutil.virtual_memory().total * 1e-9, 2)} GB",
        }
        if torch.cuda.is_available():
            self._hw_info["gpu"] = torch.cuda.get_device_name(0)

    @property
    def is_active(self):
        return self._is_active

    @staticmethod
    def _ask_permissions(timeout: int = 30):
        message = (
            "Would you like to contribute to nebullvm continuous improvement? "
            "Press enter to share the performance achieved with nebullvm and "
            '"No" to not contribute. You can find full details in the '
            '"Sharing feedback to improve nebullvm" section of the '
            'documentation. [enter/"No"] '
        )
        metadata = {}
        flag = True
        while flag:
            try:
                collect_feedback_bool = _input_with_timeout(message, timeout)
            except TimeoutError:
                collect_feedback_bool = "no"
            if len(collect_feedback_bool) == 0:
                flag = False
                metadata["allow_feedback_collection"] = True
            elif collect_feedback_bool.lower().strip() == "no":
                flag = False
                metadata["allow_feedback_collection"] = False
            message = (
                'Press enter to give your consent or type "No" to deny '
                "consent."
            )
        return metadata

    @staticmethod
    def _generate_model_id(model_name: str):
        return f"{str(uuid.uuid4())}_{hash(model_name)}"

    def start_collection(self, model: Any, framework: DeepLearningFramework):
        if not self.is_active:
            return
        if isinstance(model, str) or isinstance(model, Path):
            model_name = str(model)
        else:
            model_name = model.__class__.__name__

        self._model_id = self._generate_model_id(model_name)
        self._model_info = {
            "model_name": model_name,
            "model_size": _read_model_size(model),
            "framework": framework.value,
        }
        self._latency_dict = {}

    def store_compiler_result(
        self,
        compiler: ModelCompiler,
        q_type: Optional[QuantizationType],
        metric_drop_ths: Optional[float],
        latency: Optional[float],
        compression: str = None,
        pipeline_name: str = "onnx",
    ):
        if self._model_id is None:
            return
        q_type_key = (
            f"{q_type.value}_{metric_drop_ths}"
            if q_type is not None and metric_drop_ths is not None
            else "noopt"
        )
        if compression is not None and len(compression) > 0:
            q_type_key = compression + "_" + q_type_key
        key = pipeline_name + "_" + compiler.value
        compiler_dict = self._latency_dict.get(key, {})
        compiler_dict[q_type_key] = latency if latency else -1.0
        self._latency_dict[key] = compiler_dict

    def send_feedback(self, timeout: int = 30):
        if self._model_id is None:
            return {}
        request_body = {
            "model_id": self._model_id,
            "latency_dict": self._latency_dict,
            "hardware_setup": self._hw_info,
            "model_metadata": self._model_info,
            "nebullvm_version": VERSION,
        }
        headers = {
            "accept": "application/json",
            "Content-Type": "application/json",
        }
        url = "https://nebuly.cloud/v1/store_nebullvm_results"
        response = requests.post(
            url,
            data=json.dumps(request_body),
            headers=headers,
            timeout=timeout,
        )
        self._model_id = None
        self._latency_dict = None
        self._model_info = None
        return response


FEEDBACK_COLLECTOR = FeedbackCollector()



import warnings
from typing import Sequence, List, Tuple, Any, Union, Iterable

import numpy as np
import torch

from nebullvm.utils.onnx import convert_to_numpy


class DataManager:
    """Class for managing the user data in nebullvm.

    Attributes:
        data_reader(Sequence): Object implementing the __getitem__, the
            __len__ and the __iter__/__next__ APIs. It should read the
            user data and return tuples of tensors for feeding the models.
    """

    def __init__(self, data_reader: Sequence):
        self._data_reader = data_reader
        self._pointer = 0

    def __getitem__(self, item):
        return self._data_reader[item]

    def __len__(self):
        return len(self._data_reader)

    def __iter__(self):
        self._pointer = 0
        return self

    def __next__(self):
        if self._pointer < len(self):
            data = self[self._pointer]
            self._pointer += 1
            return data
        else:
            raise StopIteration

    def get_numpy_list(
        self, n: int, shuffle: bool = False, with_ys: bool = False
    ) -> Union[
        List[Tuple[np.ndarray, ...]], Tuple[List[Tuple[np.ndarray, ...]], List]
    ]:
        if not with_ys:
            return [
                tuple(convert_to_numpy(x) for x in tuple_)
                for tuple_ in self.get_list(n, shuffle)
            ]
        else:
            xs, ys = self.get_list(n, shuffle, with_ys=True)
            return [
                tuple(convert_to_numpy(x) for x in tuple_) for tuple_ in xs
            ], ys

    def get_list(
        self, n: int, shuffle: bool = False, with_ys: bool = False
    ) -> Union[List[Tuple[Any, ...]], Tuple[List[Tuple[Any, ...]], List]]:
        if shuffle:
            idx = np.random.choice(len(self), n, replace=n > len(self))
        else:
            idx = np.arange(0, min(n, len(self)))
            if n > len(self):
                idx = np.concatenate(
                    [
                        idx,
                        np.random.choice(
                            len(self), n - len(self), replace=True
                        ),
                    ]
                )
        if not with_ys:
            return [self[i][0] for i in idx]

        ys, xs = [], []
        for i in idx:
            x, y = self[i]
            xs.append(x)
            ys.append(y)
        return xs, ys

    @classmethod
    def from_iterable(cls, iterable: Iterable, max_length: int = 500):
        return cls([x for i, x in enumerate(iterable) if i < max_length])

    def split(self, split_pct: float, shuffle: bool = False):
        if shuffle:
            idx = np.random.choice(len(self), len(self), replace=False)
        else:
            idx = np.arange(len(self))

        n = int(round(len(idx) * split_pct))
        if n == 0 or n == len(idx):
            warnings.warn(
                "Not enough data for splitting the DataManager. "
                "An empty data-manager will be passed as result of the split."
            )
        return (
            DataManager([self[i] for i in idx[:n]]),
            DataManager([self[i] for i in idx[n:]]),
        )


class PytorchDataset(torch.utils.data.Dataset):
    def __init__(self, input_data: DataManager):
        self.data = input_data
        self.batch_size = input_data[0][0][0].shape[0]

    def __len__(self):
        return sum([batch_inputs[0].shape[0] for batch_inputs, _ in self.data])

    def __getitem__(self, idx):
        batch_idx = int(idx / self.batch_size)
        item_idx = idx % self.batch_size
        data = tuple([data[item_idx] for data in self.data[batch_idx][0]])
        return data



import os
import warnings
from pathlib import Path
from typing import List, Tuple, Optional, Callable

import numpy as np
import torch

from nebullvm.base import DeepLearningFramework, ModelParams, QuantizationType
from nebullvm.config import NVIDIA_FILENAMES, NO_COMPILER_INSTALLATION
from nebullvm.inference_learners.tensor_rt import (
    NVIDIA_INFERENCE_LEARNERS,
    NvidiaInferenceLearner,
    PytorchTensorRTInferenceLearner,
)
from nebullvm.optimizers.base import (
    BaseOptimizer,
)
from nebullvm.optimizers.quantization.tensor_rt import TensorRTCalibrator
from nebullvm.optimizers.quantization.utils import (
    check_precision,
    check_quantization,
)
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager, PytorchDataset
from nebullvm.utils.general import check_module_version
from nebullvm.utils.onnx import (
    get_input_names,
    get_output_names,
    create_model_inputs_onnx,
    run_onnx_model,
    convert_to_numpy,
)

from nebullvm.utils.torch import run_torch_model

if torch.cuda.is_available():
    try:
        import tensorrt as trt
    except ImportError:
        from nebullvm.installers.installers import install_tensor_rt

        if not NO_COMPILER_INSTALLATION:
            warnings.warn(
                "No TensorRT valid installation has been found. "
                "Trying to install it from source."
            )
            install_tensor_rt()
            import tensorrt as trt
        else:
            warnings.warn(
                "No TensorRT valid installation has been found. "
                "It won't be possible to use it in the following."
            )
    if check_module_version(torch, min_version="1.12.0"):
        try:
            import torch_tensorrt
        except ImportError:
            if not NO_COMPILER_INSTALLATION:
                from nebullvm.installers.installers import (
                    install_torch_tensor_rt,
                )

                warnings.warn(
                    "No Torch TensorRT valid installation has been found. "
                    "Trying to install it from source."
                )

                install_torch_tensor_rt()

                # Wrap import inside try/except because installation
                # may fail until wheel 1.2 will be officially out.
                try:
                    import torch_tensorrt
                except ImportError:
                    warnings.warn(
                        "Unable to install Torch TensorRT on this platform. "
                        "It won't be possible to use it in the following."
                    )
            else:
                warnings.warn(
                    "No Torch TensorRT valid installation has been found. "
                    "It won't be possible to use it in the following."
                )
    else:
        warnings.warn(
            "Torch-TensorRT can be installed only from Pytorch 1.12. "
            "Please update your Pytorch version."
        )


class TensorRTOptimizer(BaseOptimizer):
    """Class for compiling the AI models on Nvidia GPUs using TensorRT."""

    def _build_and_save_the_engine(
        self,
        engine_path: str,
        onnx_model_path: str,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation,
        quantization_type: QuantizationType = None,
        input_data: List[Tuple[np.ndarray, ...]] = None,
    ):
        # -- Build phase --
        nvidia_logger = trt.Logger(trt.Logger.WARNING)
        builder = trt.Builder(nvidia_logger)
        # create network definition
        network = builder.create_network(
            1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
        )
        # build the engine
        # TODO: setup config value for the class in a config file
        config = builder.create_builder_config()
        try:
            config.set_memory_pool_limit(trt.MemoryPoolType.WORKSPACE, 1 << 30)
        except AttributeError:
            # The method set_memory_pool_limit is not available
            # until TensorRT Release 8.4.1
            warnings.warn(
                "Cannot call method set_memory_pool_limit for TensorRT."
                "Please update TensorRT version."
            )
        if quantization_type is QuantizationType.HALF:
            config.set_flag(trt.BuilderFlag.FP16)
        elif quantization_type is QuantizationType.STATIC:
            assert input_data is not None, (
                "You need to specify the calibration data for "
                "performing static quantization."
            )
            calibrator = TensorRTCalibrator(
                batch_size=model_params.batch_size,
                input_data=input_data,
            )
            config.set_flag(trt.BuilderFlag.INT8)
            config.int8_calibrator = calibrator
        elif quantization_type is QuantizationType.DYNAMIC:
            # onnx_model_path, _ = quantize_onnx(
            #     onnx_model_path, quantization_type, input_tfms, input_data
            # )
            config.set_flag(trt.BuilderFlag.INT8)
        # import the model
        parser = trt.OnnxParser(network, nvidia_logger)
        success = parser.parse_from_file(onnx_model_path)

        if not success:
            for idx in range(parser.num_errors):
                if self.logger is not None:
                    self.logger.debug(parser.get_error(idx))
            raise ValueError(
                f"Errors occurred while processing the "
                f"ONNX file at {onnx_model_path}"
            )

        if model_params.dynamic_info is not None:
            profile = builder.create_optimization_profile()
            for input_name, input_dynamic_info, input_info in zip(
                get_input_names(onnx_model_path),
                model_params.dynamic_info.inputs,
                model_params.input_infos,
            ):
                profile.set_shape(
                    input_name,
                    (
                        min(model_params.batch_size, 1)
                        if 0 in input_dynamic_info
                        else model_params.batch_size,
                        *(
                            shape
                            if i + 1 not in input_dynamic_info
                            else (input_info.min_sizes or {}).get(i + 1, 1)
                            for i, shape in enumerate(input_info.size)
                        ),
                    ),
                    (model_params.batch_size, *input_info.size),
                    (model_params.batch_size, *input_info.size),
                )
            config.add_optimization_profile(profile)
        serialized_engine = builder.build_serialized_network(network, config)
        with open(engine_path, "wb") as f:
            f.write(serialized_engine)

    def optimize(
        self,
        model: str,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        metric: Callable = None,
        input_data: DataManager = None,
    ) -> Optional[NvidiaInferenceLearner]:
        """Optimize the input model with TensorRT.

        Args:
            model (str): Path to the saved onnx model.
            output_library (str): DL Framework the optimized model will be
                compatible with.
            model_params (ModelParams): Model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction.
            metric_drop_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with an higher drop
                will be ignored.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used.
            metric (Callable, optional): If given it should
                compute the difference between the quantized and the normal
                prediction.
            input_data (DataManager, optional): User defined data.

        Returns:
            TensorRTInferenceLearner: Model optimized with TensorRT. The model
                will have an interface in the DL library specified in
                `output_library`.
        """
        self._log(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )
        if not torch.cuda.is_available():
            raise SystemError(
                "You are trying to run an optimizer developed for NVidia gpus "
                "on a machine not connected to any GPU supporting CUDA."
            )
        check_quantization(quantization_type, metric_drop_ths)
        engine_path = Path(model).parent / NVIDIA_FILENAMES["engine"]
        if (
            metric_drop_ths is not None
            and quantization_type is QuantizationType.STATIC
        ):
            if input_data is None:
                input_data_onnx = [
                    tuple(
                        create_model_inputs_onnx(
                            model_params.batch_size, model_params.input_infos
                        )
                    )
                ]
            else:
                input_data_onnx = input_data.get_numpy_list(300, with_ys=False)
        elif (
            metric_drop_ths is not None
            and quantization_type is QuantizationType.DYNAMIC
        ):
            return None  # Dynamic quantization is not supported on tensorRT
        else:
            input_data_onnx = None
        self._build_and_save_the_engine(
            engine_path=engine_path,
            onnx_model_path=model,
            model_params=model_params,
            input_tfms=input_tfms,
            quantization_type=quantization_type,
            input_data=input_data_onnx,
        )

        learner = NVIDIA_INFERENCE_LEARNERS[output_library].from_engine_path(
            input_tfms=input_tfms,
            network_parameters=model_params,
            engine_path=engine_path,
            input_names=get_input_names(model),
            output_names=get_output_names(model),
            input_data=list(input_data.get_list(1)[0])
            if input_data is not None
            else None,
        )
        if quantization_type is not None:
            if input_data is None:
                inputs = [learner.get_inputs_example()]
                ys = None
            else:
                inputs, ys = input_data.get_numpy_list(
                    100, shuffle=True, with_ys=True
                )
            output_data = [
                tuple(
                    run_onnx_model(
                        model,
                        [convert_to_numpy(x) for x in tuple_],
                    )
                )
                for tuple_ in inputs
            ]
            is_valid = check_precision(
                learner,
                inputs,
                output_data,
                metric_drop_ths,
                metric_func=metric,
                ys=ys,
            )
            if not is_valid:
                return None
        return learner

    def optimize_from_torch(
        self,
        torch_model: torch.nn.Module,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        metric: Callable = None,
        input_data: DataManager = None,
    ) -> Optional[PytorchTensorRTInferenceLearner]:
        self._log(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )
        if not torch.cuda.is_available():
            raise SystemError(
                "You are trying to run an optimizer developed for NVidia gpus "
                "on a machine not connected to any GPU supporting CUDA."
            )

        check_quantization(quantization_type, metric_drop_ths)

        dtype = torch.float32

        if (
            metric_drop_ths is not None
            and quantization_type is QuantizationType.HALF
        ):
            dtype = torch.half
        elif (
            metric_drop_ths is not None
            and quantization_type is QuantizationType.STATIC
        ):
            dtype = torch.int8

            dataset = PytorchDataset(input_data)
            dataloader = torch.utils.data.DataLoader(
                dataset,
                batch_size=dataset.batch_size,
                shuffle=False,
                num_workers=1,
            )

            calibrator = torch_tensorrt.ptq.DataLoaderCalibrator(
                dataloader,
                use_cache=False,
                algo_type=torch_tensorrt.ptq.CalibrationAlgo.ENTROPY_CALIBRATION_2,  # noqa E501
                device=torch.device("cuda:0"),
            )
        elif (
            metric_drop_ths is not None
            and quantization_type is QuantizationType.DYNAMIC
        ):
            return None  # Dynamic quantization is not supported on tensorRT

        trt_model = torch_tensorrt.compile(
            torch_model.eval(),
            inputs=[
                torch_tensorrt.Input(
                    (model_params.batch_size, *input_info.size),
                    dtype=dtype if dtype != torch.int8 else torch.float,
                )
                for input_info in model_params.input_infos
            ],
            enabled_precisions=dtype,
            calibrator=calibrator if dtype == torch.int8 else None,
            workspace_size=1 << 22,
            device={
                "device_type": torch_tensorrt.DeviceType.GPU,
                "gpu_id": 0,
                "dla_core": 0,
                "allow_gpu_fallback": False,
                "disable_tf32": False,
            },
        )

        # Delete calibration cache
        if os.path.exists("calibration.cache"):
            os.remove("calibration.cache")

        model = PytorchTensorRTInferenceLearner(
            torch_model=trt_model,
            network_parameters=model_params,
            input_tfms=input_tfms,
            input_data=list(input_data.get_list(1)[0])
            if input_data is not None
            else None,
            dtype=dtype,
        )

        if quantization_type is not None:
            if input_data is None:
                inputs = [model.get_inputs_example()]
                ys = None
            else:
                inputs, ys = input_data.get_list(
                    100, shuffle=True, with_ys=True
                )
            output_data = [
                tuple(run_torch_model(trt_model, list(tuple_), dtype=dtype))
                for tuple_ in inputs
            ]
            is_valid = check_precision(
                model,
                inputs,
                output_data,
                metric_drop_ths,
                metric_func=metric,
                ys=ys,
            )
            if not is_valid:
                return None
        return model



import os
import shutil
import warnings
from abc import ABC
from pathlib import Path
from typing import Union, List, Generator, Tuple, Dict, Type

import numpy as np
import torch

from nebullvm.base import DeepLearningFramework, ModelParams
from nebullvm.config import ONNX_FILENAMES
from nebullvm.inference_learners.base import (
    BaseInferenceLearner,
    LearnerMetadata,
    PytorchBaseInferenceLearner,
)
from nebullvm.installers.installers import install_deepsparse
from nebullvm.transformations.base import MultiStageTransformation

try:
    from deepsparse import compile_model, cpu
except ImportError:
    import platform

    os_ = platform.system()
    if os_ != "Darwin":
        warnings.warn(
            "No deepsparse installation found. Trying to install it..."
        )
        install_deepsparse()
        from deepsparse import compile_model, cpu
    else:
        warnings.warn(
            "No valid deepsparse installation found. "
            "The compiler won't be used in the following."
        )


class DeepSparseInferenceLearner(BaseInferenceLearner, ABC):
    """Model optimized on CPU using DeepSparse. DeepSparse is an engine
    accelerating sparse computations on CPUs.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        onnx_path (str or Path): Path to the onnx model.
        input_names (List[str]): Input names used when the onnx model
            was produced.
        output_names (List[str]): Output names used when the onnx model
            was produced.
    """

    def __init__(
        self,
        onnx_path: Union[str, Path],
        input_names: List[str],
        output_names: List[str],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.onnx_path = self._store_file(onnx_path)

        # Compile model
        cores_per_socket, _, _ = cpu.cpu_details()
        # Define the number of cores to use, by default it will make use of
        # all physical cores on the system
        num_cores = cores_per_socket
        batch_size = kwargs["network_parameters"].batch_size
        self.engine = compile_model(onnx_path, batch_size, num_cores)

        self.input_names = input_names
        self.output_names = output_names

    def save(self, path: Union[str, Path], **kwargs):
        """Save the model.

        Args:
            path (Path or str): Path to the directory where the model will
                be stored.
            kwargs (Dict): Dictionary of key-value pairs that will be saved in
                the model metadata file.
        """
        metadata = LearnerMetadata.from_model(
            self,
            input_names=self.input_names,
            output_names=self.output_names,
            **kwargs,
        )
        metadata.save(path)

        shutil.copy(
            self.onnx_path,
            Path(path) / ONNX_FILENAMES["model_name"],
        )

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        """Load the model.

        Args:
            path (Path or str): Path to the directory where the model is
                stored.
            kwargs (Dict): Dictionary of additional arguments for consistency
                with other Learners.

        Returns:
            DeepSparseInferenceLearner: The optimized model.
        """
        if len(kwargs) > 0:
            warnings.warn(
                f"No extra keywords expected for the load method. "
                f"Got {kwargs}."
            )
        onnx_path = os.path.join(str(path), ONNX_FILENAMES["model_name"])
        metadata = LearnerMetadata.read(path)
        input_tfms = metadata.input_tfms
        if input_tfms is not None:
            input_tfms = MultiStageTransformation.from_dict(
                metadata.input_tfms
            )
        return cls(
            input_tfms=input_tfms,
            network_parameters=ModelParams(**metadata.network_parameters),
            onnx_path=onnx_path,
            input_names=metadata["input_names"],
            output_names=metadata["output_names"],
        )

    def _predict_arrays(self, input_arrays: Generator[np.ndarray, None, None]):
        inputs = [array for array in input_arrays]
        outputs = self.engine(inputs)
        return outputs


class PytorchDeepSparseInferenceLearner(
    DeepSparseInferenceLearner, PytorchBaseInferenceLearner
):
    """Model optimized on CPU using DeepSparse. DeepSparse is an engine
    accelerating sparse computations on CPUs.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        onnx_path (str or Path): Path to the onnx model.
        input_names (List[str]): Input names used when the onnx model
            was produced.
        output_names (List[str]): Output names used when the onnx model
            was produced.
    """

    def run(self, *input_tensors: torch.Tensor) -> Tuple[torch.Tensor]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[Tensor]): Input tensors belonging to the same
                batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        input_arrays = (
            input_tensor.cpu().detach().numpy()
            for input_tensor in input_tensors
        )
        outputs = self._predict_arrays(input_arrays)
        return tuple(torch.from_numpy(output) for output in outputs)


DEEPSPARSE_INFERENCE_LEARNERS: Dict[
    DeepLearningFramework, Type[DeepSparseInferenceLearner]
] = {DeepLearningFramework.PYTORCH: PytorchDeepSparseInferenceLearner}



from tempfile import TemporaryDirectory

import pytest

from nebullvm.api.huggingface import HuggingFaceInferenceLearner
from nebullvm.base import DeepLearningFramework
from nebullvm.optimizers.extra import HuggingFaceOptimizer
from nebullvm.optimizers.tests.utils import get_huggingface_model


@pytest.mark.parametrize(
    ("output_library",),
    [(DeepLearningFramework.PYTORCH,)],
)
def test_huggingface(output_library: DeepLearningFramework):
    with TemporaryDirectory() as tmp_dir:
        (
            model,
            model_params,
            output_structure,
            input_names,
            output_type,
            input_data,
        ) = get_huggingface_model(tmp_dir, output_library)

        optimizer = HuggingFaceOptimizer({})
        model = optimizer.optimize(
            model, output_library, model_params, input_data=input_data
        )

        model = HuggingFaceInferenceLearner(
            core_inference_learner=model,
            output_structure=output_structure,
            input_names=input_names,
            output_type=output_type,
        )

        assert isinstance(model, HuggingFaceInferenceLearner)

        # Test save and load functions
        model.save(tmp_dir)
        loaded_model = HuggingFaceInferenceLearner.load(tmp_dir)
        assert isinstance(loaded_model, HuggingFaceInferenceLearner)

        inputs_example = list(model.get_inputs_example())
        res = model.predict(*inputs_example)
        assert res is not None



import os
import platform
import subprocess
from pathlib import Path
import sys

import cpuinfo
import torch

from nebullvm.utils.general import check_module_version


def _get_cpu_arch():
    arch = cpuinfo.get_cpu_info()["arch"].lower()
    if "x86" in arch:
        return "x86"
    else:
        return "arm"


def _get_os():
    return platform.system()


def install_tvm(working_dir: str = None):
    """Helper function for installing ApacheTVM.

    This function needs some prerequisites for running, as a valid `git`
    installation and having MacOS or a Linux-distribution as OS.

    Args:
        working_dir (str, optional): The directory where the tvm repo will be
            cloned and installed.
    """
    path = Path(__file__).parent
    # install pre-requisites
    installation_file_prerequisites = str(
        path / "install_tvm_prerequisites.sh"
    )
    subprocess.run(
        ["bash", installation_file_prerequisites],
        cwd=working_dir or Path.home(),
    )
    installation_file = str(path / "install_tvm.sh")
    hardware_config = _get_cpu_arch()
    if torch.cuda.is_available():
        hardware_config = f"{hardware_config}_cuda"
    env_dict = {
        "CONFIG_PATH": str(
            path / f"tvm_installers/{hardware_config}/config.cmake"
        ),
        **dict(os.environ.copy()),
    }
    subprocess.run(
        ["bash", installation_file],
        cwd=working_dir or Path.home(),
        env=env_dict,
    )


def install_bladedisc():
    """Helper function for installing BladeDisc."""
    has_cuda = False
    if torch.cuda.is_available():
        has_cuda = True

    path = Path(__file__).parent
    installation_file = str(path / "install_bladedisc.sh")
    subprocess.Popen(["bash", installation_file, str(has_cuda).lower()])


def install_torch_tensor_rt():
    """Helper function for installing Torch-TensorRT.

    The function will install the software only if a cuda driver is available.
    """
    if not torch.cuda.is_available():
        raise RuntimeError(
            "Torch-TensorRT can run just on Nvidia machines. "
            "No available cuda driver has been found."
        )
    elif not check_module_version(torch, min_version="1.12.0"):
        raise RuntimeError(
            "Torch-TensorRT can be installed only from Pytorch 1.12. "
            "Please update your Pytorch version."
        )

    # Verify that TensorRT is installed, otherwise install it
    try:
        import tensorrt  # noqa F401
    except ImportError:
        install_tensor_rt()

    # # Will work when Torch-TensorRT v1.2 will be available
    # cmd = [
    #     "pip3",
    #     "install",
    #     "torch-tensorrt",
    #     "-f",
    #     "https://github.com/pytorch/TensorRT/releases",
    # ]
    # subprocess.run(cmd)

    # Install Torch-TensorRT from alpha wheel
    wheels_dict = {
        "37": "1nTEk1gyOx87hapRuORik9OhjudXKvvnG",
        "38": "1IYcMFf9eeESOsvOIZgE2E2NJ9tjRvfri",
        "39": "15vMu3dzd3-hRUnIiIIbagvU-UJftM9i1",
        "310": "1sVODbKNd66h0W86T9VQ6QXS1t2ZUQIdr",
    }

    python_version = str(sys.version_info.major) + str(sys.version_info.minor)
    tensor_rt_wheel = (
        f"torch_tensorrt-1.2.0a0-cp{python_version}-cp"
        f"{python_version + 'm' if python_version == '37' else python_version}"
        f"-linux_x86_64.whl"
    )

    wheel_id = wheels_dict.get(python_version)
    if wheel_id is not None:
        cmd = (
            f"wget --no-check-certificate https://drive.google.com/uc?export="
            f"download&id={wheel_id} -O {tensor_rt_wheel}"
        )

        subprocess.run(cmd.split())

        cmd = [
            "pip",
            "install",
            "./" + tensor_rt_wheel,
        ]
        subprocess.run(cmd)

        os.remove("./" + tensor_rt_wheel)


def install_tensor_rt():
    """Helper function for installing TensorRT.

    The function will install the software only if a cuda driver is available.
    """
    if not torch.cuda.is_available():
        raise RuntimeError(
            "TensorRT can run just on Nvidia machines. "
            "No available cuda driver has been found."
        )
    path = Path(__file__).parent
    installation_file = str(path / "install_tensor_rt.sh")
    subprocess.run(["bash", installation_file])


def install_openvino(with_optimization: bool = True):
    """Helper function for installing the OpenVino compiler.

    This function just works on intel machines.

    Args:
        with_optimization (bool): Flag for installing the full openvino engine
            or limiting the installation to the tools need for inference
            models.
    """
    processor = cpuinfo.get_cpu_info()["brand_raw"].lower()
    if "intel" not in processor:
        raise RuntimeError(
            f"Openvino can run just on Intel machines. "
            f"You are trying to install it on {processor}"
        )
    openvino_version = "openvino-dev" if with_optimization else "openvino"
    cmd = ["pip3", "install", f"{openvino_version}[onnx]"]
    subprocess.run(cmd)
    cmd = ["pip3", "install", "numpy>=1.20,<1.23"]
    subprocess.run(cmd)


def install_onnxruntime():
    """Helper function for installing the right version of onnxruntime."""
    distribution_name = "onnxruntime"
    if torch.cuda.is_available():
        distribution_name = f"{distribution_name}-gpu"
    if _get_os() == "Darwin" and _get_cpu_arch() == "arm":
        cmd = ["conda", "install", "-y", distribution_name]
    else:
        cmd = ["pip3", "install", distribution_name]
    subprocess.run(cmd)
    # install requirements for onnxruntime.transformers
    cmd = ["pip3", "install", "coloredlogs", "sympy"]
    subprocess.run(cmd)


def install_deepsparse():
    """Helper function for installing DeepSparse."""
    cmd = ["pip3", "install", "deepsparse"]
    subprocess.run(cmd)



import copy
import logging
from abc import ABC, abstractmethod
from logging import Logger
from typing import Dict, List, Any, Callable, Tuple, Optional

import cpuinfo
import numpy as np
import tensorflow as tf
import torch.nn
from tqdm import tqdm

from nebullvm.base import (
    DeepLearningFramework,
    ModelParams,
    QuantizationType,
    ModelCompiler,
    OptimizationTime,
)
from nebullvm.compressors.base import BaseCompressor
from nebullvm.compressors.intel import TorchIntelPruningCompressor
from nebullvm.compressors.sparseml import SparseMLCompressor
from nebullvm.inference_learners import (
    BaseInferenceLearner,
    PytorchBaseInferenceLearner,
)
from nebullvm.measure import compute_optimized_running_time
from nebullvm.optimizers import (
    BaseOptimizer,
    ApacheTVMOptimizer,
    COMPILER_TO_OPTIMIZER_MAP,
    DeepSparseOptimizer,
)
from nebullvm.optimizers.blade_disc import BladeDISCOptimizer
from nebullvm.optimizers.pytorch import PytorchBackendOptimizer
from nebullvm.optimizers.tensor_rt import TensorRTOptimizer
from nebullvm.optimizers.tensorflow import TensorflowBackendOptimizer
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.compilers import (
    tvm_is_available,
    select_compilers_from_hardware_onnx,
    deepsparse_is_available,
    bladedisc_is_available,
    torch_tensorrt_is_available,
)
from nebullvm.utils.data import DataManager
from nebullvm.utils.feedback_collector import FEEDBACK_COLLECTOR


class Step(ABC):
    """Fundamental building block for the Pipeline.

    Attributes:
        logger (Logger, optional): Logger defined by the user.
    """

    def __init__(self, logger: Logger = None):
        self._logger = logger

    @abstractmethod
    def run(self, *args, **kwargs) -> Dict:
        """Run the pipeline step."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def name(self):
        raise NotImplementedError()

    def _log_info(self, text: str):
        if self._logger is None:
            logging.info(text)
        else:
            self._logger.info(text)

    def _log_warning(self, text: str):
        if self._logger is None:
            logging.warning(text)
        else:
            self._logger.warning(text)


class CompressorStep(Step, ABC):
    """Object managing the Compressor step in the Pipeline. This step manages
    all the defined Compressor objects available considering the data given
    by the user.

    Attributes:
        config_file (str, optional): The configuration file containing the
            configuration parameter for each Compressor. The config_file is
            a YAML file having as main keywords the Compressor names and as
            values dictionaries containing the specific parameters for the
            related Compressor object.
        logger (Logger, optional): Logger defined by the user.
    """

    def __init__(self, config_file: str = None, logger: Logger = None):
        super().__init__(logger)
        self._config_file = config_file

    def run(
        self,
        model: Any = None,
        input_data: DataManager = None,
        metric_drop_ths: float = None,
        metric: Callable = None,
        **kwargs,
    ) -> Dict:
        """Run the CompressorStep.

        Args:
            model (Any): Model to be compressed.
            input_data (DataManager): Data to be used for compressing the
                model.
            metric_drop_ths: Maximum reduction in the selected metric accepted.
                No model with an higher error will be accepted. Note that the
                maximum error is modified and then propagated to the next
                steps.
            metric (Callable): Metric to be used for estimating the error
                due to the compression.
            kwargs (Dict): Keyword arguments propagated to the next step.
        """
        compressor_dict = self._get_compressors()
        self._log_info(f"Compressions: {tuple(compressor_dict.keys())}")
        models = {"no_compression": (copy.deepcopy(model), metric_drop_ths)}
        train_input_data, eval_input_data = input_data.split(0.8)
        for technique, compressor in tqdm(compressor_dict.items()):
            try:
                compressed_model, ths = compressor.compress(
                    model,
                    train_input_data,
                    eval_input_data,
                    metric_drop_ths,
                    metric,
                )
                models[technique] = (compressed_model, ths)
            except Exception as ex:
                self._log_warning(
                    f"Error during compression {technique}. Got error {ex}. "
                    f"The compression technique will be skipped. "
                    f"Please consult the documentation for further info or "
                    f"open an issue on GitHub for receiving assistance."
                )
        return {
            "models": models,
            "input_data": eval_input_data,
            "metric": metric,
            **kwargs,
        }

    @abstractmethod
    def _get_compressors(self) -> Dict[str, BaseCompressor]:
        raise NotImplementedError()

    @property
    def name(self):
        return "compression_step"


class TorchCompressorStep(CompressorStep):
    """Object managing the Compressor step in the Pipeline for PyTorch models.
    This step manages all the defined Compressor objects available considering
    the data given by the user.

    At the current state this step supports pruning with SparseML and (just on
    intel devices) pruning with the IntelNeuralCompressor.

    Attributes:
        config_file (str, optional): The configuration file containing the
            configuration parameter for each Compressor. The config_file is
            a YAML file having as main keywords the Compressor names and as
            values dictionaries containing the specific parameters for the
            related Compressor object.
        logger (Logger, optional): Logger defined by the user.
    """

    def _get_compressors(self) -> Dict[str, BaseCompressor]:
        compressors = {
            "sparseml": SparseMLCompressor(config_file=self._config_file)
        }
        # TODO: Reactivate the intel-neural-compressor when properly tested
        if False and "intel" in cpuinfo.get_cpu_info()["brand_raw"].lower():
            compressors["intel_pruning"] = TorchIntelPruningCompressor(
                config_file=self._config_file
            )
        return compressors


class NoCompressionStep(Step):
    """Step to be used when no compression is required.

    Attributes:
        logger (Logger, optional): Logger defined by the user.
    """

    def run(
        self, model: Any, metric_drop_ths: Optional[float], **kwargs
    ) -> Dict:
        return {
            "models": {"no_compression": (model, metric_drop_ths)},
            **kwargs,
        }

    @property
    def name(self):
        return "no_compression"


class OptimizerStep(Step, ABC):
    """Object managing the Optimizers in the pipeline step. All available
    optimizers are run on the model given as input and a list of tuples
    (optimized_model, latency) is given as output.

    Attributes:
        logger (Logger, optional): Logger defined by the user.
    """

    def run(
        self,
        models: Dict[str, Tuple[Any, Optional[float]]] = None,
        output_library: DeepLearningFramework = None,
        model_params: ModelParams = None,
        input_tfms: MultiStageTransformation = None,
        metric: Optional[Callable] = None,
        input_data: Optional[DataManager] = None,
        ignore_compilers: List[ModelCompiler] = None,
        optimization_time: OptimizationTime = None,
        pipeline_name: str = None,
        **kwargs,
    ) -> Dict:
        """Run the OptimizerStep for all the available compilers.

        Args:
            models (Dict): Dictionary of models produced by the CompressorStep.
                For each model produced by the previous step, the updated
                metric_drop_ths (i.e. the error allowed on the model) is
                given together with the model. Keys represent the compression
                technique used for obtaining the model.
            output_library (DeepLearningFramework): The target framework.
            model_params (ModelParams): The model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction.
            metric (Callable): Metric to be used for estimating the error
                due to the compression.
            input_data (DataManager): Input data to be used for optimizing the
                model.
            ignore_compilers (List): List of compilers to be ignored.
            optimization_time (OptimizationTime): The optimization time mode.
                It can be either 'constrained' or 'unconstrained'. For
                'unconstrained' optimization all the compilers are re-used on
                the different framework interfaces, even if the model has
                already been compiled with the same compiler on another
                framework interface.
            pipeline_name (str): Name of the pipeline.
            kwargs (Dict): Extra keywords that will be ignored.
        """

        optimizers = self._get_optimizers(ignore_compilers)
        self._log_info(
            f"Optimizations: "
            f"{tuple(compiler.value for compiler in optimizers.keys())}"
        )
        optimized_models = []

        for prev_tech, (model, metric_drop_ths) in tqdm(models.items()):
            self._log_info(f"Optimizing output of {prev_tech}")
            if model is None:
                continue
            if metric_drop_ths is not None:
                q_types = [
                    None,
                    QuantizationType.DYNAMIC,
                    QuantizationType.HALF,
                ]
                if input_data is not None:
                    q_types.append(QuantizationType.STATIC)
            else:
                q_types = [None]
            for compiler, optimizer in tqdm(optimizers.items()):
                for q_type in q_types:
                    try:
                        optimized_model = self._run_optimizer(
                            optimizer,
                            model,
                            output_library,
                            model_params,
                            input_tfms.copy(),
                            metric_drop_ths,
                            q_type,
                            metric,
                            input_data,
                        )
                        if optimized_model is not None:
                            latency = compute_optimized_running_time(
                                optimized_model
                            )
                        else:
                            latency = np.inf
                        optimized_models.append((optimized_model, latency))
                        if (
                            compiler not in ignore_compilers
                            and optimization_time
                            is OptimizationTime.CONSTRAINED
                        ):
                            ignore_compilers.append(compiler)
                        FEEDBACK_COLLECTOR.store_compiler_result(
                            compiler=compiler,
                            q_type=q_type,
                            metric_drop_ths=metric_drop_ths,
                            latency=latency,
                            compression=prev_tech,
                            pipeline_name=pipeline_name,
                        )
                    except Exception as ex:
                        self._log_warning(
                            f"Compilation failed with {output_library.value} "
                            f"interface of {compiler}. Got error {ex}. "
                            f"If possible the compilation will be re-scheduled"
                            f" with another interface. Please consult the "
                            f"documentation for further info or open an issue "
                            f"on GitHub for receiving assistance."
                        )
                        FEEDBACK_COLLECTOR.store_compiler_result(
                            compiler=compiler,
                            q_type=q_type,
                            metric_drop_ths=metric_drop_ths,
                            latency=None,
                            compression=prev_tech,
                            pipeline_name=pipeline_name,
                        )

        return {
            "optimized_models": optimized_models,
            "ignore_compilers": ignore_compilers,
        }

    @property
    def name(self):
        return "optimizer_step"

    @abstractmethod
    def _run_optimizer(
        self,
        optimizer,
        model: Any,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        metric: Callable = None,
        input_data: DataManager = None,
    ) -> BaseInferenceLearner:
        raise NotImplementedError()

    @abstractmethod
    def _get_optimizers(
        self, ignore_compilers: List[ModelCompiler]
    ) -> Dict[ModelCompiler, BaseOptimizer]:
        raise NotImplementedError()


class TorchOptimizerStep(OptimizerStep):
    """Object managing the Optimizers in the pipeline step supporting PyTorch
    as compiler interface. All available optimizers are run on the model given
    as input and a list of tuples (optimized_model, latency) is given as
    output.

    Attributes:
        logger (Logger, optional): Logger defined by the user.
    """

    def _get_optimizers(
        self, ignore_compilers: List[ModelCompiler]
    ) -> Dict[ModelCompiler, BaseOptimizer]:
        optimizers = {
            ModelCompiler.TORCHSCRIPT: PytorchBackendOptimizer(
                logger=self._logger
            ),
        }
        if (
            tvm_is_available()
            and ModelCompiler.APACHE_TVM not in ignore_compilers
        ):
            optimizers[ModelCompiler.APACHE_TVM] = ApacheTVMOptimizer(
                logger=self._logger
            )
        if (
            deepsparse_is_available()
            and ModelCompiler.DEEPSPARSE not in ignore_compilers
        ):
            optimizers[ModelCompiler.DEEPSPARSE] = DeepSparseOptimizer(
                logger=self._logger
            )
        if (
            bladedisc_is_available()
            and ModelCompiler.BLADEDISC not in ignore_compilers
        ):
            optimizers[ModelCompiler.BLADEDISC] = BladeDISCOptimizer(
                logger=self._logger
            )
        if (
            torch_tensorrt_is_available()
            and ModelCompiler.TENSOR_RT not in ignore_compilers
        ):
            optimizers[ModelCompiler.TENSOR_RT] = TensorRTOptimizer(
                logger=self._logger
            )
        return optimizers

    def _run_optimizer(
        self,
        optimizer,
        model: Any,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        metric: Callable = None,
        input_data: DataManager = None,
    ) -> PytorchBaseInferenceLearner:
        if hasattr(optimizer, "optimize_from_torch"):
            optimized_model = optimizer.optimize_from_torch(
                torch_model=model,
                model_params=model_params,
                metric_drop_ths=metric_drop_ths
                if quantization_type is not None
                else None,
                metric=metric,
                quantization_type=quantization_type,
                input_tfms=input_tfms,
                input_data=input_data,
            )
        else:
            optimized_model = optimizer.optimize(
                model=model,
                output_library=output_library,
                model_params=model_params,
                metric_drop_ths=metric_drop_ths
                if quantization_type is not None
                else None,
                metric=metric,
                quantization_type=quantization_type,
                input_tfms=input_tfms,
                input_data=input_data,
            )
        return optimized_model


class TFOptimizerStep(OptimizerStep):
    """Object managing the Optimizers in the pipeline step supporting
    TensorFlow as compiler interface. All available optimizers are run on
    the model given as input and a list of tuples (optimized_model, latency)
    is given as output.

    Attributes:
        logger (Logger, optional): Logger defined by the user.
    """

    def _get_optimizers(
        self, ignore_compilers: List[ModelCompiler]
    ) -> Dict[ModelCompiler, BaseOptimizer]:
        optimizers = {
            ModelCompiler.TFLITE: TensorflowBackendOptimizer(
                logger=self._logger
            )
        }
        return optimizers

    def _run_optimizer(
        self,
        optimizer,
        model: Any,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        metric: Callable = None,
        input_data: DataManager = None,
    ) -> PytorchBaseInferenceLearner:
        if hasattr(optimizer, "optimize_from_tf"):
            optimized_model = optimizer.optimize_from_tf(
                torch_model=model,
                model_params=model_params,
                metric_drop_ths=metric_drop_ths
                if quantization_type is not None
                else None,
                metric=metric,
                quantization_type=quantization_type,
                input_tfms=input_tfms,
                input_data=input_data,
            )
        else:
            optimized_model = optimizer.optimize(
                model=model,
                output_library=output_library,
                model_params=model_params,
                metric_drop_ths=metric_drop_ths
                if quantization_type is not None
                else None,
                metric=metric,
                quantization_type=quantization_type,
                input_tfms=input_tfms,
                input_data=input_data,
            )
        return optimized_model


class OnnxOptimizerStep(OptimizerStep):
    """Object managing the Optimizers in the pipeline step supporting ONNX
    as compiler interface. All available optimizers are run on the model given
    as input and a list of tuples (optimized_model, latency) is given as
    output.

    Attributes:
        logger (Logger, optional): Logger defined by the user.
    """

    def _get_optimizers(
        self, ignore_compilers: List[ModelCompiler]
    ) -> Dict[ModelCompiler, BaseOptimizer]:
        compilers = select_compilers_from_hardware_onnx()
        optimizers = {
            compiler: COMPILER_TO_OPTIMIZER_MAP[compiler](self._logger)
            for compiler in compilers
            if compiler not in ignore_compilers
        }
        return optimizers

    def _run_optimizer(
        self,
        optimizer,
        model: Any,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        metric_drop_ths: float = None,
        quantization_type: QuantizationType = None,
        metric: Callable = None,
        input_data: DataManager = None,
    ) -> PytorchBaseInferenceLearner:
        optimized_model = optimizer.optimize(
            model=model,
            output_library=output_library,
            model_params=model_params,
            metric_drop_ths=metric_drop_ths
            if quantization_type is not None
            else None,
            metric=metric,
            quantization_type=quantization_type,
            input_tfms=input_tfms,
            input_data=input_data,
        )
        return optimized_model


class Pipeline(Step):
    """Pipeline object.

    A Pipeline is a list of steps executed sequentially, where each step
    takes as input the output of the previous one.

    Attributes:
        pipeline_name: str,
        steps (List): List of Steps composing the pipeline.
        logger (Logger): Logger defined by the user.
    """

    def __init__(
        self, pipeline_name: str, steps: List[Step], logger: Logger = None
    ):
        super().__init__(logger)
        self._name = pipeline_name
        self._steps = steps

    def run(self, **kwargs) -> Dict:
        self._log_info(f"Running pipeline: {self.name}")
        kwargs["pipeline_name"] = self.name.split("_")[0]
        for step in self._steps:
            self._log_info(f"Running step: {step.name}")
            kwargs = step.run(**kwargs)
        return kwargs

    @property
    def name(self):
        return self._name


def _get_compressor_step(
    model: Any,
    optimization_time: OptimizationTime,
    config_file: Optional[str],
    metric_drop_ths: Optional[float],
    metric: Optional[Callable],
    logger: Optional[Logger],
) -> Step:
    if optimization_time is OptimizationTime.CONSTRAINED:
        return NoCompressionStep(logger=logger)
    if metric_drop_ths is None or metric is None:
        return NoCompressionStep(logger=logger)
    elif isinstance(model, torch.nn.Module):
        return TorchCompressorStep(config_file=config_file, logger=logger)
    else:  # default is NoCompression
        return NoCompressionStep(logger=logger)


def _get_optimizer_step(
    model: Any,
    logger: Optional[Logger],
) -> Step:
    if isinstance(model, torch.nn.Module):
        return TorchOptimizerStep(logger=logger)
    elif isinstance(model, tf.Module):
        return TFOptimizerStep(logger=logger)
    else:
        return OnnxOptimizerStep(logger=logger)


def _get_pipeline_name(model: Any):
    if isinstance(model, torch.nn.Module):
        return "pytorch_pipeline"
    elif isinstance(model, tf.Module):
        return "tensorflow_pipeline"
    else:
        return "onnx_pipeline"


def build_pipeline_from_model(
    model: Any,
    optimization_time: OptimizationTime,
    metric_drop_ths: Optional[float],
    metric: Optional[Callable],
    config_file: Optional[str],
    logger: Logger = None,
) -> Pipeline:
    """Function for building a pipeline from a model and user-defined
    parameters

    Args:
        model (Any): The input model.
        optimization_time (OptimizationTime): The optimization time mode.
            It can be either 'constrained' or 'unconstrained'. For
            'constrained' mode just compilers and precision reduction
            techniques are used (no compression). 'Unconstrained' optimization
            allows the usage of more time consuming techniques as pruning and
            distillation.
        metric_drop_ths (float, optional): Maximum reduction in the
            selected metric accepted. No model with an higher error will be
            accepted, i.e. all optimized model having a larger error respect to
            the original one will be discarded, without even considering their
            possible speed-up.
        metric (Callable): Metric to be used for estimating the error
            due to the optimization techniques.
        config_file (str, optional): Configuration file containing the
            parameters needed for defining the CompressionStep in the pipeline.
        logger (Logger, optional): Logger defined by the user.
    """
    compressor_step = _get_compressor_step(
        model, optimization_time, config_file, metric_drop_ths, metric, logger
    )
    optimizer_step = _get_optimizer_step(model, logger)
    pipeline = Pipeline(
        pipeline_name=_get_pipeline_name(model),
        logger=logger,
        steps=[compressor_step, optimizer_step],
    )
    return pipeline



import json
import shutil
import warnings
from abc import ABC
from pathlib import Path
from typing import Dict, Union, Type, Generator, Tuple, List

import cpuinfo
import numpy as np
import tensorflow as tf
import torch

from nebullvm.base import ModelParams, DeepLearningFramework
from nebullvm.config import (
    OPENVINO_FILENAMES,
    NO_COMPILER_INSTALLATION,
)
from nebullvm.inference_learners.base import (
    BaseInferenceLearner,
    LearnerMetadata,
    PytorchBaseInferenceLearner,
    TensorflowBaseInferenceLearner,
    NumpyBaseInferenceLearner,
)
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager

try:
    from openvino.runtime import Core, Model, CompiledModel, InferRequest
except ImportError:
    if (
        "intel" in cpuinfo.get_cpu_info()["brand_raw"].lower()
        and not NO_COMPILER_INSTALLATION
    ):
        warnings.warn(
            "No valid OpenVino installation has been found. "
            "Trying to re-install it from source."
        )
        from nebullvm.installers.installers import install_openvino

        install_openvino(with_optimization=True)
        from openvino.runtime import Core, Model, CompiledModel, InferRequest
    else:
        warnings.warn(
            "No Openvino library detected. "
            "The Openvino Inference learner should not be used."
        )
        Model = CompiledModel = InferRequest = object


class OpenVinoInferenceLearner(BaseInferenceLearner, ABC):
    """Model optimized using ApacheTVM.

    The class cannot be directly instantiated, but implements all the core
    methods needed for using ApacheTVM at inference time.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        exec_network (any): The graph executor. This is the
            central component in the OpenVino optimized model execution.
        input_keys (List): Keys associated to the inputs.
        output_keys (List): Keys associated to the outputs.
        description_file (str): File containing a description of the optimized
            model.
        weights_file (str): File containing the model weights.
    """

    def __init__(
        self,
        model: Model,
        compiled_model: CompiledModel,
        infer_request: InferRequest,
        input_keys: List,
        output_keys: List,
        description_file: str,
        weights_file: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.model = model
        self.compiled_model = compiled_model
        self.infer_request = infer_request
        self.input_keys = input_keys
        self.output_keys = output_keys
        self.description_file = self._store_file(description_file)
        self.weights_file = self._store_file(weights_file)

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        """Load the model.

        Args:
            path (Path or str): Path to the directory where the model is
                stored.
            kwargs (Dict): Dictionary of additional arguments for the
                `from_model_name` class method.

        Returns:
            OpenVinoInferenceLearner: The optimized model.
        """
        path = Path(path)
        with open(path / OPENVINO_FILENAMES["metadata"], "r") as fin:
            metadata = json.load(fin)
        metadata.update(kwargs)
        metadata["network_parameters"] = ModelParams(
            **metadata["network_parameters"]
        )
        input_tfms = metadata.get("input_tfms")
        if input_tfms is not None:
            metadata["input_tfms"] = MultiStageTransformation.from_dict(
                input_tfms
            )
        model_name = str(path / OPENVINO_FILENAMES["description_file"])
        model_weights = str(path / OPENVINO_FILENAMES["weights"])
        return cls.from_model_name(
            model_name=model_name, model_weights=model_weights, **metadata
        )

    @classmethod
    def from_model_name(
        cls,
        network_parameters: ModelParams,
        model_name: str,
        model_weights: str,
        input_tfms: MultiStageTransformation = None,
        input_data: DataManager = None,
        **kwargs,
    ):
        """Build the optimized model from the network description and its
        weights.

        Args:
            network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
            model_name (str): File containing a description of the optimized
                model.
            model_weights (str): File containing the model weights.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction.
            input_data (DataManager, optional): User defined data.
        """
        if len(kwargs) > 0:
            warnings.warn(f"Found extra parameters: {kwargs}")

        core = Core()
        model = core.read_model(model=model_name, weights=model_weights)
        compiled_model = core.compile_model(model=model, device_name="CPU")
        infer_request = compiled_model.create_infer_request()

        input_keys = list(
            map(lambda obj: obj.get_any_name(), compiled_model.inputs)
        )
        output_keys = list(
            map(lambda obj: obj.get_any_name(), compiled_model.outputs)
        )

        return cls(
            model,
            compiled_model,
            infer_request,
            input_keys,
            output_keys,
            input_tfms=input_tfms,
            network_parameters=network_parameters,
            description_file=model_name,
            weights_file=model_weights,
            input_data=input_data,
        )

    def _rebuild_network(self, input_shapes: Dict):
        # get the input name and input shape info from model
        model_inputs_info = {
            item.get_any_name(): tuple(item.shape)
            for item in self.model.inputs
        }
        if all(
            input_shape == model_inputs_info[input_name]
            for input_name, input_shape in input_shapes.items()
        ):
            # If the new input shapes is equal to the previous one do nothing.
            return

        core = Core()
        model = core.read_model(
            model=self.description_file, weights=self.weights_file
        )
        model.reshape(input_shapes)
        compiled_model = core.compile_model(model=model, device_name="CPU")
        infer_request = compiled_model.create_infer_request()

        self.model = model
        self.compiled_model = compiled_model
        self.infer_request = infer_request

    def _get_metadata(self, **kwargs) -> LearnerMetadata:
        # metadata = {
        #     key: self.__dict__[key] for key in ("input_keys", "output_keys")
        # }
        metadata = {}
        metadata.update(kwargs)
        return LearnerMetadata.from_model(self, **metadata)

    def save(self, path: Union[str, Path], **kwargs):
        """Save the model.

        Args:
            path (Path or str): Path to the directory where the model will
                be stored.
            kwargs (Dict): Dictionary of key-value pairs that will be saved in
                the model metadata file.
        """
        path = Path(path)
        path.mkdir(exist_ok=True)
        metadata = self._get_metadata(**kwargs)
        with open(path / OPENVINO_FILENAMES["metadata"], "w") as fout:
            json.dump(metadata.to_dict(), fout)

        shutil.copy(
            self.description_file,
            path / OPENVINO_FILENAMES["description_file"],
        )
        shutil.copy(self.weights_file, path / OPENVINO_FILENAMES["weights"])

    def _predict_array(
        self,
        input_arrays: Generator[np.ndarray, None, None],
        input_shapes: Generator[Tuple[int, ...], None, None] = None,
    ) -> Generator[np.ndarray, None, None]:
        if input_shapes is not None:
            input_shapes_dict = {
                name: size for name, size in zip(self.input_keys, input_shapes)
            }
            self._rebuild_network(input_shapes_dict)

        results = self.infer_request.infer(
            inputs={
                input_key: input_array
                for input_key, input_array in zip(
                    self.input_keys, input_arrays
                )
            }
        )
        results = {
            output_key.get_any_name(): output_arr
            for output_key, output_arr in results.items()
        }

        return (results[output_key] for output_key in self.output_keys)


class PytorchOpenVinoInferenceLearner(
    OpenVinoInferenceLearner, PytorchBaseInferenceLearner
):
    """Model optimized using ApacheTVM with a Pytorch interface.

    This class can be used exactly in the same way as a pytorch Module object.
    At prediction time it takes as input pytorch tensors given as positional
    arguments.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        exec_network (any): The graph executor. This is the
            central component in the OpenVino optimized model execution.
        input_keys (List): Keys associated to the inputs.
        output_keys (List): Keys associated to the outputs.
        description_file (str): File containing a description of the optimized
            model.
        weights_file (str): File containing the model weights.
    """

    def run(self, *input_tensors: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[Tensor]): Input tensors belonging to the same
                batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        input_arrays = (
            input_tensor.cpu().detach().numpy()
            for input_tensor in input_tensors
        )
        extra_kwargs = {}
        if self.network_parameters.dynamic_info is not None:
            extra_kwargs["input_shapes"] = (
                tuple(input_tensor.size()) for input_tensor in input_tensors
            )
        output_arrays = self._predict_array(input_arrays, **extra_kwargs)
        return tuple(
            torch.from_numpy(output_array) for output_array in output_arrays
        )


class TensorflowOpenVinoInferenceLearner(
    OpenVinoInferenceLearner, TensorflowBaseInferenceLearner
):
    """Model optimized using ApacheTVM with a tensorflow interface.

    This class can be used exactly in the same way as a tf.Module or
    keras.Model object.
    At prediction time it takes as input tensorflow tensors given as positional
    arguments.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        exec_network (any): The graph executor. This is the
            central component in the OpenVino optimized model execution.
        input_keys (List): Keys associated to the inputs.
        output_keys (List): Keys associated to the outputs.
        description_file (str): File containing a description of the optimized
            model.
        weights_file (str): File containing the model weights.
    """

    def run(self, *input_tensors: tf.Tensor) -> Tuple[tf.Tensor, ...]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[Tensor]): Input tensors belonging to the same
                batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        input_arrays = (input_tensor.numpy() for input_tensor in input_tensors)
        extra_kwargs = {}
        if self.network_parameters.dynamic_info is not None:
            extra_kwargs["input_shapes"] = (
                tuple(input_tensor.shape) for input_tensor in input_tensors
            )
        output_arrays = self._predict_array(input_arrays, **extra_kwargs)
        # noinspection PyTypeChecker
        return tuple(
            tf.convert_to_tensor(output_array)
            for output_array in output_arrays
        )


class NumpyOpenVinoInferenceLearner(
    OpenVinoInferenceLearner, NumpyBaseInferenceLearner
):
    """Model optimized using ApacheTVM with a numpy interface.

    This class can be used exactly in the same way as a sklearn or
    numpy-based model.
    At prediction time it takes as input numpy arrays given as positional
    arguments.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        exec_network (any): The graph executor. This is the
            central component in the OpenVino optimized model execution.
        input_keys (List): Keys associated to the inputs.
        output_keys (List): Keys associated to the outputs.
        description_file (str): File containing a description of the optimized
            model.
        weights_file (str): File containing the model weights.
    """

    def run(self, *input_tensors: np.ndarray) -> Tuple[np.ndarray, ...]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[np.ndarray]): Input tensors belonging to
                the same batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[np.ndarray]: Output tensors. Note that the output tensors
                does not correspond to the prediction on the input tensors
                with a 1 to 1 mapping. In fact the output tensors are produced
                as the multiple-output of the model given a (multi-) tensor
                input.
        """
        input_arrays = (input_tensor for input_tensor in input_tensors)
        extra_kwargs = {}
        if self.network_parameters.dynamic_info is not None:
            extra_kwargs["input_shapes"] = (
                tuple(input_tensor.shape) for input_tensor in input_tensors
            )
        output_arrays = self._predict_array(input_arrays, **extra_kwargs)
        return tuple(output_arrays)


OPENVINO_INFERENCE_LEARNERS: Dict[
    DeepLearningFramework, Type[OpenVinoInferenceLearner]
] = {
    DeepLearningFramework.PYTORCH: PytorchOpenVinoInferenceLearner,
    DeepLearningFramework.TENSORFLOW: TensorflowOpenVinoInferenceLearner,
    DeepLearningFramework.NUMPY: NumpyOpenVinoInferenceLearner,
}
